@@ -15,7 +15,8 @@ from typing import Tuple, Dict, List, Optional, Iterable, Mapping
 import click
 from defusedxml import minidom
 
-from eodatasets3 import DatasetAssembler
+from eodatasets3 import DatasetPrepare
+from eodatasets3.properties import Eo3Interface
 from eodatasets3.ui import PathPath
 
 # Static namespace to generate uuids for datacube indexing
@@ -159,7 +160,7 @@ def process_user_product_metadata(contents: str) -> Dict:
     }
 
 
-def _get_stable_id(p: DatasetAssembler) -> uuid.UUID:
+def _get_stable_id(p: Eo3Interface) -> uuid.UUID:
     # These should have been extracted from our metadata files!
     producer = p.producer
     product_name = p.properties["sentinel:product_name"]
@@ -171,20 +172,20 @@ def _get_stable_id(p: DatasetAssembler) -> uuid.UUID:
 
 
 def prepare_and_write(
-    ds_path: Path,
+    dataset_location: Path,
     output_yaml: Path,
     producer: str,
 ) -> Tuple[uuid.UUID, Path]:
-    with DatasetAssembler(
+    with DatasetPrepare(
         metadata_path=output_yaml,
-        dataset_location=ds_path,
+        dataset_location=dataset_location,
     ) as p:
         p.properties["odc:producer"] = producer
 
         if producer == "esa.int":
-            jp2_offsets = _extract_esa_fields(ds_path, p)
+            jp2_offsets = _extract_esa_fields(dataset_location, p)
         elif producer == "sinergise.com":
-            jp2_offsets = _extract_sinergise_fields(ds_path, p)
+            jp2_offsets = _extract_sinergise_fields(dataset_location.parent, p)
         else:
             raise NotImplementedError(
                 f"Unknown s2 producer {producer}. Expected 'sinergise.com' or 'esa.int'"
@@ -235,7 +236,7 @@ def _get_platform_name(p: Mapping) -> str:
     return platform_name
 
 
-def _extract_sinergise_fields(path: Path, p: DatasetAssembler) -> Iterable[Path]:
+def _extract_sinergise_fields(path: Path, p: DatasetPrepare) -> Iterable[Path]:
     """Extract Sinergise metadata and return list of image offsets"""
     product_info_path = path / "productInfo.json"
     metadata_xml_path = path / "metadata.xml"
@@ -247,14 +248,14 @@ def _extract_sinergise_fields(path: Path, p: DatasetAssembler) -> Iterable[Path]
         )
 
     p.properties.update(process_sinergise_product_info(product_info_path))
-    p.add_accessory_file("metadata:sinergise_product_info", product_info_path)
+    p.note_accessory_file("metadata:sinergise_product_info", product_info_path)
 
     p.properties.update(process_tile_metadata(metadata_xml_path.read_text()))
-    p.add_accessory_file("metadata:s2_tile", metadata_xml_path)
+    p.note_accessory_file("metadata:s2_tile", metadata_xml_path)
 
     # TODO: sinergise folders could `process_datastrip_metadata()` in an outer directory?
 
-    return path.glob("*.jp2")
+    return list(path.glob("*.jp2"))
 
 
 def _extract_esa_fields(dataset, p) -> Iterable[Path]:
@@ -274,17 +275,17 @@ def _extract_esa_fields(dataset, p) -> Iterable[Path]:
         p.properties.update(
             process_datastrip_metadata(z.read(datastrip_md).decode("utf-8"))
         )
-        p.add_accessory_file("metadata:s2_datastrip", datastrip_md)
+        p.note_accessory_file("metadata:s2_datastrip", datastrip_md)
 
         tile_md = one("MTD_TL.xml")
         p.properties.update(process_tile_metadata(z.read(tile_md).decode("utf-8")))
-        p.add_accessory_file("metadata:s2_tile", tile_md)
+        p.note_accessory_file("metadata:s2_tile", tile_md)
 
         user_product_md = one("MTD_MSIL1C.xml")
         p.properties.update(
             process_user_product_metadata(z.read(user_product_md).decode("utf-8"))
         )
-        p.add_accessory_file("metadata:s2_user_product", user_product_md)
+        p.note_accessory_file("metadata:s2_user_product", user_product_md)
 
         return [Path(p) for p in z.namelist() if "IMG_DATA" in p and p.endswith(".jp2")]
 
@@ -340,8 +341,8 @@ def main(
             *(
                 (
                     "sinergise.com",
-                    # Input is the folder
-                    p.parent,
+                    # Input folder: our indexed dataset location is the metadata file.
+                    (p.parent / f"{p.parent.stem}.odc-metadata.yaml"),
                     # Output is an inner metadata file, with the same name as the folder (usually S2A....).
                     (p.parent / f"{p.parent.stem}.odc-metadata.yaml"),
                 )
@@ -350,7 +351,7 @@ def main(
             *(
                 (
                     "esa.int",
-                    # Input is zip file
+                    # Input is zip file, so the zip should be our location.
                     p,
                     # Output is a sibling file with metadata suffix.
                     p.with_suffix(".odc-metadata.yaml"),
@@ -366,7 +367,7 @@ def main(
 
         for producer, ds_path, output_yaml in in_out_paths:
             if output_base:
-                output_yaml = output_base / output_yaml.name
+                output_yaml = output_base.absolute() / output_yaml.name
 
             if output_yaml.exists():
                 if not overwrite_existing:
@@ -375,8 +376,8 @@ def main(
 
                 logging.info("Output exists: overwriting %s", output_yaml)
 
-            uuid, path = prepare_and_write(ds_path, output_yaml, producer)
-            logging.info("Wrote dataset %s to %s", uuid, path)
+            uuid_, path = prepare_and_write(ds_path, output_yaml, producer)
+            logging.info("Wrote dataset %s to %s", uuid_, path)
 
         sys.exit(0)
 
