@@ -129,6 +129,16 @@ def degrees_type(value):
     return value
 
 
+def identifier_type(v: str):
+    v = v.replace("-", "_")
+    if not v.isidentifier() or not v.islower():
+        warnings.warn(
+            f"{v!r} is expected to be an identifier "
+            "(alphanumeric with underscores, typically lowercase)"
+        )
+    return v
+
+
 def producer_check(value):
     if "." not in value:
         warnings.warn(
@@ -209,6 +219,8 @@ _GQA_FMASK_PROPS = {
     "gqa:abs_xy": float,
     "gqa:abs_y": float,
     "gqa:cep90": float,
+    "gqa:error_message": None,
+    "gqa:final_gcp_count": int,
     "gqa:iterative_mean_x": float,
     "gqa:iterative_mean_xy": float,
     "gqa:iterative_mean_y": float,
@@ -218,6 +230,7 @@ _GQA_FMASK_PROPS = {
     "gqa:mean_x": float,
     "gqa:mean_xy": float,
     "gqa:mean_y": float,
+    "gqa:ref_source": None,
     "gqa:stddev_x": float,
     "gqa:stddev_xy": float,
     "gqa:stddev_y": float,
@@ -243,6 +256,13 @@ _LANDSAT_EXTENDED_PROPS = {
     "landsat:processing_software_version": None,
     "landsat:scan_gap_interpolation": float,
     "landsat:station_id": None,
+    # Landsat USGS Properties
+    "landsat:rmse": None,
+    "landsat:rmse_x": None,
+    "landsat:rmse_y": None,
+    "landsat:wrs_type": None,
+    "landsat:correction": None,
+    "landsat:cloud_cover_land": None,
 }
 
 _SENTINEL_EXTENDED_PROPS = {
@@ -261,13 +281,31 @@ _SENTINEL_EXTENDED_PROPS = {
 }
 
 
-class StacPropertyView(collections.abc.MutableMapping):
+_STAC_MISC_PROPS = {
+    "providers": None,  # https://github.com/radiantearth/stac-spec/blob/master/item-spec/common-metadata.md#provider,
+    # Projection extension
+    "proj:epsg": int,
+    "proj:shape": None,
+    "proj:transform": None,
+}
+
+
+class Eo3Dict(collections.abc.MutableMapping):
+    """
+    This acts like a dictionary, but will normalise known properties (consistent
+    case, types etc) and warn about common mistakes.
+
+    It wraps an inner dictionary. By default it will normalise the fields in
+    the input dictionary on creation, but you can disable this with `normalise_input=False`.
+    """
+
     # Every property we've seen or dealt with so far. Feel free to expand with abandon...
     # This is to minimise minor typos, case differences, etc, which plagued previous systems.
     # Keep sorted.
-    KNOWN_STAC_PROPERTIES: Mapping[str, Optional[NormaliseValueFn]] = {
+    KNOWN_PROPERTIES: Mapping[str, Optional[NormaliseValueFn]] = {
         "datetime": datetime_type,
         "dea:dataset_maturity": of_enum_type(("final", "interim", "nrt"), lower=True),
+        "dea:product_maturity": of_enum_type(("stable", "provisional"), lower=True),
         "dtr:end_datetime": datetime_type,
         "dtr:start_datetime": datetime_type,
         "eo:azimuth": float,
@@ -296,16 +334,25 @@ class StacPropertyView(collections.abc.MutableMapping):
         "odc:processing_datetime": datetime_type,
         "odc:producer": producer_check,
         "odc:product": None,
-        "odc:product_family": None,
+        "odc:product_family": identifier_type,
         "odc:region_code": None,
         **_LANDSAT_EXTENDED_PROPS,
         **_GQA_FMASK_PROPS,
         **_SENTINEL_EXTENDED_PROPS,
+        **_STAC_MISC_PROPS,
     }
 
-    def __init__(self, properties=None) -> None:
-        self._props = properties or {}
+    # For backwards compatibility, in case users are extending at runtime.
+    KNOWN_STAC_PROPERTIES = KNOWN_PROPERTIES
 
+    def __init__(self, properties: Mapping = None, normalise_input=True) -> None:
+        if properties is None:
+            properties = {}
+        self._props = properties
+        # We normalise the properties they gave us.
+        if normalise_input:
+            for key in list(self._props):
+                self.normalise_and_set(key, self._props[key], expect_override=True)
         self._finished_init_ = True
 
     def __setattr__(self, name: str, value: Any) -> None:
@@ -331,6 +378,9 @@ class StacPropertyView(collections.abc.MutableMapping):
     def __delitem__(self, name: str) -> None:
         del self._props[name]
 
+    def __repr__(self) -> str:
+        return f"{self.__class__.__name__}({self._props!r})"
+
     def __setitem__(self, key, value):
         self.normalise_and_set(
             key,
@@ -339,12 +389,20 @@ class StacPropertyView(collections.abc.MutableMapping):
             allow_override=True,
         )
 
-    def normalise_and_set(self, key, value, allow_override=True):
+    def normalise_and_set(self, key, value, allow_override=True, expect_override=False):
         """
-        Normalise the given value if it's a known key (eg. dates should be dates),
-        and set it on the given dictionary.
+        Set a property with the usual normalisation.
+
+        This has some options that are not available on normal dictionary item
+        setting (``self[key] = val``)
+
+        The default behaviour of this class is very conservative in order to catch common errors
+        of users. You can loosen the settings here.
+
+        :argument allow_override: Is it okay to overwrite an existing value? (if not, error will be thrown)
+        :argument expect_override: We expect to overwrite a property, so don't produce a warning or error.
         """
-        if key not in self.KNOWN_STAC_PROPERTIES:
+        if key not in self.KNOWN_PROPERTIES:
             warnings.warn(
                 f"Unknown Stac property {key!r}. "
                 f"If this is valid property, please tell us on Github here so we can add it: "
@@ -352,7 +410,7 @@ class StacPropertyView(collections.abc.MutableMapping):
             )
 
         if value is not None:
-            normalise = self.KNOWN_STAC_PROPERTIES.get(key)
+            normalise = self.KNOWN_PROPERTIES.get(key)
             if normalise:
                 value = normalise(value)
                 # If the normaliser has extracted extra properties, we'll get two return values.
@@ -365,7 +423,7 @@ class StacPropertyView(collections.abc.MutableMapping):
                             )
                         self.normalise_and_set(k, v, allow_override=allow_override)
 
-        if key in self._props and value != self[key]:
+        if key in self._props and value != self[key] and (not expect_override):
             message = (
                 f"Overriding property {key!r} " f"(from {self[key]!r} to {value!r})"
             )
@@ -380,20 +438,47 @@ class StacPropertyView(collections.abc.MutableMapping):
         return nest_properties(self._props)
 
 
+class StacPropertyView(Eo3Dict):
+    """
+    Backwards compatibility class name. Deprecated.
+
+    Use the identical 'Eo3Dict' instead.
+
+    These were called  "StacProperties" in Stac 0.6, but many of them have
+    changed in newer versions and we're sticking to the old names for consistency
+    and backwards-compatibility. So they're now EO3 Properties.
+
+    (The eo3-to-stac tool to convert EO3 properties to real Stac properties.)
+    """
+
+    def __init__(self, properties=None) -> None:
+        super().__init__(properties)
+        warnings.warn(
+            "The class name 'StacPropertyView' is deprecated as it's misleading. "
+            "Please change your import to the (identical) 'Eo3Dict'.",
+            category=DeprecationWarning,
+        )
+
+
 class PropertyOverrideWarning(UserWarning):
     """A warning that a property was set twice with different values."""
 
     ...
 
 
-class EoFields:
+class Eo3Interface:
     """
-    Convenient access fields for the most common/essential properties in datasets
+    These are convenience properties for common metadata fields. They are available
+    on DatasetAssemblers and within other naming APIs.
+
+    (This is abstract. If you want one of these of your own, you probably want to create
+    an :class:`eodatasets3.DatasetDoc`)
+
     """
 
     @property
     @abstractmethod
-    def properties(self) -> StacPropertyView:
+    def properties(self) -> Eo3Dict:
         raise NotImplementedError
 
     @property
@@ -401,12 +486,12 @@ class EoFields:
         """
         Unique name of the specific platform the instrument is attached to.
 
-        For satellites this would be the name of the satellite (e.g., landsat-8, sentinel-2A),
+        For satellites this would be the name of the satellite (e.g., ``landsat-8``, ``sentinel-2a``),
         whereas for drones this would be a unique name for the drone.
 
-        In derivative products, multiple platforms can be specified with a comma "landsat5,landsat7".
+        In derivative products, multiple platforms can be specified with a comma: ``landsat-5,landsat-7``.
 
-        Shorthand for 'eo:platform' property
+        Shorthand for ``eo:platform`` property
         """
         return self.properties.get("eo:platform")
 
@@ -435,7 +520,7 @@ class EoFields:
         """
         Name of instrument or sensor used (e.g., MODIS, ASTER, OLI, Canon F-1).
 
-        Shorthand for 'eo:instrument' property
+        Shorthand for ``eo:instrument`` property
         """
         return self.properties.get("eo:instrument")
 
@@ -446,7 +531,7 @@ class EoFields:
     @property
     def constellation(self) -> str:
         """
-        Constellation. Eg 'sentinel-2".
+        Constellation. Eg ``sentinel-2``.
         """
         return self.properties.get("eo:constellation")
 
@@ -457,7 +542,7 @@ class EoFields:
     @property
     def product_name(self) -> Optional[str]:
         """
-        The ODC product
+        The ODC product name
         """
         return self.properties.get("odc:product")
 
@@ -470,9 +555,9 @@ class EoFields:
         """
         Organisation that produced the data.
 
-        eg. usgs.gov or ga.gov.au
+        eg. ``usgs.gov`` or ``ga.gov.au``
 
-        Shorthand for 'odc:producer' property
+        Shorthand for ``odc:producer`` property
         """
         return self.properties.get("odc:producer")
 
@@ -485,10 +570,10 @@ class EoFields:
         """
         An optional date range for the dataset.
 
-        The `datetime` is still mandatory when this is set.
+        The ``datetime`` is still mandatory when this is set.
 
         This field is a shorthand for reading/setting the datetime-range
-        stac extension properties: 'dtr:start_datetime' and 'dtr:end_datetime'
+        stac 0.6 extension properties: ``dtr:start_datetime`` and ``dtr:end_datetime``
         """
         return (
             self.properties.get("dtr:start_datetime"),
@@ -502,28 +587,16 @@ class EoFields:
         self.properties["dtr:start_datetime"] = start
         self.properties["dtr:end_datetime"] = end
 
-    @property
-    def datetime(self) -> datetime:
-        """
-        The searchable date and time of the assets. (Default to UTC if not specified)
-        """
-        return self.properties.get("datetime")
-
-    @datetime.setter
-    def datetime(self, val: datetime) -> datetime:
-        self.properties["datetime"] = val
-
     @property
     def processed(self) -> datetime:
-        """
-        When the dataset was processed (Default to UTC if not specified)
+        """When the dataset was created (Defaults to UTC if not specified)
 
-        Shorthand for the 'odc:processing_datetime' field
+        Shorthand for the ``odc:processing_datetime`` field
         """
         return self.properties.get("odc:processing_datetime")
 
     @processed.setter
-    def processed(self, value):
+    def processed(self, value: Union[str, datetime]):
         self.properties["odc:processing_datetime"] = value
 
     def processed_now(self):
@@ -537,7 +610,7 @@ class EoFields:
         """
         The version of the dataset.
 
-        Typically digits separated by a dot. Eg. '1.0.0'
+        Typically digits separated by a dot. Eg. `1.0.0`
 
         The first digit is usually the collection number for
         this 'producer' organisation, such as USGS Collection 1 or
@@ -549,11 +622,14 @@ class EoFields:
     def collection_number(self) -> int:
         """
         The version of the collection.
-        Eg:
-          metadata:
-            product_family: wofs
-            dataset_version: 1.6.0
-            collection_number: 3
+
+        Eg.::
+
+           metadata:
+             product_family: wofs
+             dataset_version: 1.6.0
+             collection_number: 3
+
         """
         return self.properties.get("odc:collection_number")
 
@@ -576,17 +652,17 @@ class EoFields:
     @property
     def product_family(self) -> str:
         """
-        The identifier for this "family" of products, such as 'ard', 'level1` or 'fc'.
+        The identifier for this "family" of products, such as ``ard``, ``level1`` or ``fc``.
         It's used for grouping similar products together.
 
         They products in a family are usually produced the same way but have small variations:
         they come from different sensors, or are written in different projections, etc.
 
-        'ard' family of products: 'ls7_ard', 'ls5_ard' ....
+        ``ard`` family of products: ``ls7_ard``, ``ls5_ard`` ....
 
-        On older versions of opendatacube this was called "product_type".
+        On older versions of Open Data Cube this was called ``product_type``.
 
-        Shorthand for 'odc:product_family' property.
+        Shorthand for ``odc:product_family`` property.
         """
         return self.properties.get("odc:product_family")
 
@@ -594,6 +670,10 @@ class EoFields:
     def product_family(self, value):
         self.properties["odc:product_family"] = value
 
+    @product_family.deleter
+    def product_family(self):
+        del self.properties["odc:product_family"]
+
     @property
     def region_code(self) -> Optional[str]:
         """
@@ -603,11 +683,11 @@ class EoFields:
 
         It's generally treated as an opaque string to group datasets and process as stacks.
 
-        For Landsat products it's the concatenated '{path}{row}' (both numbers formatted to three digits).
+        For Landsat products it's the concatenated ``{path}{row}`` (both numbers formatted to three digits).
 
         For Sentinel 2, it's the MGRS grid (TODO presumably?).
 
-        Shorthand for 'odc:region_code' property.
+        Shorthand for ``odc:region_code`` property.
         """
         return self.properties.get("odc:region_code")
 
@@ -621,7 +701,7 @@ class EoFields:
         The dataset maturity. The same data may be processed multiple times -- becoming more
         mature -- as new ancillary data becomes available.
 
-        Typical values (from least to most mature): "nrt", "interim", "final"
+        Typical values (from least to most mature): ``nrt`` (near real time), ``interim``, ``final``
         """
         return self.properties.get("dea:dataset_maturity")
 
@@ -629,6 +709,33 @@ class EoFields:
     def maturity(self, value):
         self.properties["dea:dataset_maturity"] = value
 
+    @property
+    def product_maturity(self) -> str:
+        """
+        Classification: is this a 'provisional' or 'stable' release of the product?
+        """
+        return self.properties.get("dea:product_maturity")
+
+    @product_maturity.setter
+    def product_maturity(self, value):
+        self.properties["dea:product_maturity"] = value
+
+    # Note that giving a method the name 'datetime' will override the 'datetime' type
+    # for class-level declarations (ie, for any types on functions!)
+    # So we make an alias:
+    from datetime import datetime as datetime_
+
+    @property
+    def datetime(self) -> datetime_:
+        """
+        The searchable date and time of the assets. (Default to UTC if not specified)
+        """
+        return self.properties.get("datetime")
+
+    @datetime.setter
+    def datetime(self, val: datetime_):
+        self.properties["datetime"] = val
+
 
 def _github_suggest_new_property_url(key: str, value: object) -> str:
     """Get a URL to create a Github issue suggesting new properties to be added."""
