@@ -24,7 +24,7 @@ from shapely.geometry.base import BaseGeometry
 from eodatasets3.model import (
     DatasetDoc,
     ODC_DATASET_SCHEMA_URL,
-    StacPropertyView,
+    Eo3Dict,
 )
 from eodatasets3.properties import FileFormat
 
@@ -42,7 +42,7 @@ def _uuid_representer(dumper, data):
     return dumper.represent_scalar("tag:yaml.org,2002:str", f"{data}")
 
 
-def represent_datetime(self, data: datetime):
+def _represent_datetime(self, data: datetime):
     """
     The default Ruamel representer strips 'Z' suffixes for UTC.
 
@@ -57,11 +57,11 @@ def represent_datetime(self, data: datetime):
     return self.represent_scalar("tag:yaml.org,2002:timestamp", value)
 
 
-def represent_numpy_datetime(self, data: numpy.datetime64):
-    return represent_datetime(self, data.astype("M8[ms]").tolist())
+def _represent_numpy_datetime(self, data: numpy.datetime64):
+    return _represent_datetime(self, data.astype("M8[ms]").tolist())
 
 
-def represent_paths(self, data: PurePath):
+def _represent_paths(self, data: PurePath):
     return Representer.represent_str(self, data.as_posix())
 
 
@@ -70,8 +70,8 @@ def _init_yaml() -> YAML:
 
     yaml.representer.add_representer(FileFormat, _format_representer)
     yaml.representer.add_multi_representer(UUID, _uuid_representer)
-    yaml.representer.add_representer(datetime, represent_datetime)
-    yaml.representer.add_multi_representer(PurePath, represent_paths)
+    yaml.representer.add_representer(datetime, _represent_datetime)
+    yaml.representer.add_multi_representer(PurePath, _represent_paths)
 
     # WAGL spits out many numpy primitives in docs.
     yaml.representer.add_representer(numpy.int8, Representer.represent_int)
@@ -85,7 +85,7 @@ def _init_yaml() -> YAML:
     yaml.representer.add_representer(numpy.float32, Representer.represent_float)
     yaml.representer.add_representer(numpy.float64, Representer.represent_float)
     yaml.representer.add_representer(numpy.ndarray, Representer.represent_list)
-    yaml.representer.add_representer(numpy.datetime64, represent_numpy_datetime)
+    yaml.representer.add_representer(numpy.datetime64, _represent_numpy_datetime)
 
     # Match yamllint default expectations. (Explicit start/end are recommended to tell if a file is cut off)
     yaml.width = 80
@@ -126,6 +126,13 @@ def loads_yaml(stream: Union[Text, IO]) -> Iterable[Dict]:
 
 
 def from_path(path: Path, skip_validation=False) -> DatasetDoc:
+    """
+    Parse an EO3 document from a filesystem path
+
+    :param path: Filesystem path
+    :param skip_validation: Optionally disable validation (it's faster, but I hope your
+            doc is structured correctly)
+    """
     if path.suffix.lower() not in (".yaml", ".yml"):
         raise ValueError(f"Unexpected file type {path.suffix}. Expected yaml")
 
@@ -186,10 +193,14 @@ METADATA_TYPE_SCHEMA = _load_schema_validator(
 
 def from_doc(doc: Dict, skip_validation=False) -> DatasetDoc:
     """
-    Convert a document to a dataset.
+    Parse a dictionary into an EO3 dataset.
 
     By default it will validate it against the schema, which will result in far more
     useful error messages if fields are missing.
+
+    :param doc: A dictionary, such as is returned from yaml.load or json.load
+    :param skip_validation: Optionally disable validation (it's faster, but I hope your
+            doc is structured correctly)
     """
 
     if not skip_validation:
@@ -198,15 +209,18 @@ def from_doc(doc: Dict, skip_validation=False) -> DatasetDoc:
     # TODO: stable cattrs (<1.0) balks at the $schema variable.
     doc = doc.copy()
     del doc["$schema"]
+    location = doc.pop("location", None)
+    if location:
+        doc["locations"] = [location]
 
     c = cattr.Converter()
     c.register_structure_hook(uuid.UUID, _structure_as_uuid)
     c.register_structure_hook(BaseGeometry, _structure_as_shape)
-    c.register_structure_hook(StacPropertyView, _structure_as_stac_props)
+    c.register_structure_hook(Eo3Dict, _structure_as_stac_props)
 
     c.register_structure_hook(Affine, _structure_as_affine)
 
-    c.register_unstructure_hook(StacPropertyView, _unstructure_as_stac_props)
+    c.register_unstructure_hook(Eo3Dict, _unstructure_as_stac_props)
     return c.structure(doc, DatasetDoc)
 
 
@@ -215,7 +229,8 @@ def _structure_as_uuid(d, t):
 
 
 def _structure_as_stac_props(d, t):
-    return StacPropertyView(d)
+    # We don't normalise properties as we want it to reflect the original file.
+    return Eo3Dict(d, normalise_input=False)
 
 
 def _structure_as_affine(d: Tuple, t):
@@ -232,7 +247,7 @@ def _structure_as_affine(d: Tuple, t):
     return Affine(*d)
 
 
-def _unstructure_as_stac_props(v: StacPropertyView):
+def _unstructure_as_stac_props(v: Eo3Dict):
     return v._props
 
 
@@ -245,7 +260,7 @@ def to_doc(d: DatasetDoc) -> Dict:
     Serialise a DatasetDoc to a dict
 
     If you plan to write this out as a yaml file on disk, you're
-    better off with `to_formatted_doc()`.
+    better off with one of our formatted writers: :func:`.to_stream`, :func:`.to_path`.
     """
     doc = attr.asdict(
         d,
@@ -263,6 +278,10 @@ def to_doc(d: DatasetDoc) -> Dict:
         doc["geometry"] = shapely.geometry.mapping(d.geometry)
     doc["id"] = str(d.id)
     doc["properties"] = dict(d.properties)
+
+    if len(doc.get("locations", [])) == 1:
+        doc["location"] = doc.pop("locations")[0]
+
     return doc
 
 
@@ -278,6 +297,24 @@ def to_formatted_doc(d: DatasetDoc) -> CommentedMap:
     return doc
 
 
+def to_path(path: Path, *ds: DatasetDoc):
+    """
+    Output dataset(s) as a formatted YAML to a local path
+
+    (multiple datasets will result in a multi-document yaml file)
+    """
+    dump_yaml(path, *(to_formatted_doc(d) for d in ds))
+
+
+def to_stream(stream, *ds: DatasetDoc):
+    """
+    Output dataset(s) as a formatted YAML to an output stream
+
+    (multiple datasets will result in a multi-document yaml file)
+    """
+    dumps_yaml(stream, *(to_formatted_doc(d) for d in ds))
+
+
 def _stac_key_order(key: str):
     """All keys in alphabetical order, but unprefixed keys first."""
     if ":" in key:
@@ -316,6 +353,8 @@ _EO3_PROPERTY_ORDER = [
     "id",
     "label",
     "product",
+    "location",
+    "locations",
     "crs",
     "geometry",
     "grids",
@@ -323,8 +362,6 @@ _EO3_PROPERTY_ORDER = [
     "measurements",
     "accessories",
     "lineage",
-    "location",
-    "locations",
 ]
 
 
