@@ -4,9 +4,11 @@ Validate ODC dataset documents
 import collections
 import enum
 import math
+import multiprocessing
 import sys
 from datetime import datetime
 from pathlib import Path
+from textwrap import indent
 from typing import (
     List,
     Counter,
@@ -18,21 +20,27 @@ from typing import (
     Sequence,
     Iterable,
 )
+from urllib.parse import urljoin
+from urllib.request import urlopen
 
 import attr
+import ciso8601
 import click
 import numpy as np
 import rasterio
 from boltons.iterutils import get_path
 from click import style, echo, secho
+from datacube import Datacube
+from datacube.utils import changes, read_documents, is_url, InvalidDocException
+from datacube.utils.documents import load_documents
 from rasterio import DatasetReader
 from rasterio.crs import CRS
 from rasterio.errors import CRSError
 from shapely.validation import explain_validity
 
-from eodatasets3 import serialise, model
+from eodatasets3 import serialise, model, utils
 from eodatasets3.model import DatasetDoc
-from eodatasets3.ui import PathPath, is_absolute, uri_resolve, bool_style
+from eodatasets3.ui import is_absolute, uri_resolve, bool_style
 from eodatasets3.utils import default_utc, EO3_SCHEMA
 
 
@@ -69,15 +77,15 @@ SUFFIX_KINDS = {
 DOC_TYPE_SUFFIXES = {v: k for k, v in SUFFIX_KINDS.items()}
 
 
-def filename_doc_kind(path: Path) -> Optional["DocKind"]:
+def filename_doc_kind(path: Union[str, Path]) -> Optional["DocKind"]:
     """
     Get the expected file type for the given filename.
 
     Returns None if it does not follow any naming conventions.
 
-    >>> filename_doc_kind(Path('LC8_2014.odc-metadata.yaml')).name
+    >>> filename_doc_kind('LC8_2014.odc-metadata.yaml').name
     'dataset'
-    >>> filename_doc_kind(Path('/tmp/something/water_bodies.odc-metadata.yaml.gz')).name
+    >>> filename_doc_kind('/tmp/something/water_bodies.odc-metadata.yaml.gz').name
     'dataset'
     >>> filename_doc_kind(Path('/tmp/something/ls8_fc.odc-product.yaml')).name
     'product'
@@ -88,7 +96,7 @@ def filename_doc_kind(path: Path) -> Optional["DocKind"]:
     >>> filename_doc_kind(Path('/tmp/something/some_other_file.yaml'))
     """
 
-    for suffix in reversed(path.suffixes):
+    for suffix in reversed(Path(path).suffixes):
         suffix = suffix.lower()
         if suffix in SUFFIX_KINDS:
             return SUFFIX_KINDS[suffix]
@@ -153,6 +161,7 @@ def validate_dataset(
     thorough: bool = False,
     readable_location: Union[str, Path] = None,
     expect_extra_measurements: bool = False,
+    expect_geometry: bool = True,
 ) -> ValidationMessages:
     """
     Validate a a dataset document, optionally against the given product.
@@ -203,7 +212,7 @@ def validate_dataset(
     if not dataset.product.href:
         _info("product_href", "A url (href) is recommended for products")
 
-    yield from _validate_geo(dataset)
+    yield from _validate_geo(dataset, expect_geometry=expect_geometry)
 
     # Note that a dataset may have no measurements (eg. telemetry data).
     # (TODO: a stricter mode for when we know we should have geo and measurement info)
@@ -480,72 +489,72 @@ class ExpectedMeasurement:
 
 
 def validate_paths(
-    paths: List[Path],
+    paths: List[str],
     thorough: bool = False,
     expect_extra_measurements: bool = False,
-) -> Generator[Tuple[Path, int, List[ValidationMessage]], None, None]:
+    product_definitions: Dict[str, Dict] = None,
+) -> Generator[Tuple[str, List[ValidationMessage]], None, None]:
     """Validate the list of paths. Product documents can be specified before their datasets."""
-    products: Dict[str, Dict] = {}
 
-    for path, was_specified_by_user in expand_directories(paths):
-        with path.open("r") as f:
-            for i, doc in enumerate(serialise.loads_yaml(f)):
-                messages = []
-                kind = filename_doc_kind(path)
-                if kind is None:
-                    kind = guess_kind_from_contents(doc)
-                    if kind and (kind in DOC_TYPE_SUFFIXES):
-                        # It looks like an ODC doc but doesn't have the standard suffix.
-                        messages.append(
-                            _warning(
-                                "missing_suffix",
-                                f"Document looks like a {kind.name} but does not have "
-                                f'filename extension "{DOC_TYPE_SUFFIXES[kind]}{_readable_doc_extension(path)}"',
-                            )
-                        )
+    products = dict(product_definitions or {})
 
-                if kind == DocKind.product:
-                    messages.extend(validate_product(doc))
-                    if "name" in doc:
-                        products[doc["name"]] = doc
-                elif kind == DocKind.dataset:
-                    messages.extend(
-                        validate_eo3_doc(
-                            doc, path, products, thorough, expect_extra_measurements
-                        )
+    for url, doc, was_specified_by_user in read_paths(paths):
+        messages = []
+        kind = filename_doc_kind(url)
+        if kind is None:
+            kind = guess_kind_from_contents(doc)
+            if kind and (kind in DOC_TYPE_SUFFIXES):
+                # It looks like an ODC doc but doesn't have the standard suffix.
+                messages.append(
+                    _warning(
+                        "missing_suffix",
+                        f"Document looks like a {kind.name} but does not have "
+                        f'filename extension "{DOC_TYPE_SUFFIXES[kind]}{_readable_doc_extension(url)}"',
                     )
-                elif kind == DocKind.metadata_type:
-                    messages.extend(validate_metadata_type(doc))
-                # Otherwise it's a file we don't support.
-                # If the user gave us the path explicitly, it seems to be an error.
-                # (if they didn't -- it was found via scanning directories -- we don't care.)
-                elif was_specified_by_user:
-                    if kind is None:
-                        raise ValueError(f"Unknown document type for path {path}")
-                    else:
-                        raise NotImplementedError(
-                            f"Cannot currently validate {kind.name} files"
-                        )
-                else:
-                    # Not a doc type we recognise, and the user didn't specify it. Skip it.
-                    continue
+                )
 
-                yield path, i, messages
+        if kind == DocKind.product:
+            messages.extend(validate_product(doc))
+            if "name" in doc:
+                products[doc["name"]] = doc
+        elif kind == DocKind.dataset:
+            messages.extend(
+                validate_eo3_doc(
+                    doc, url, products, thorough, expect_extra_measurements
+                )
+            )
+        elif kind == DocKind.metadata_type:
+            messages.extend(validate_metadata_type(doc))
+        # Otherwise it's a file we don't support.
+        # If the user gave us the path explicitly, it seems to be an error.
+        # (if they didn't -- it was found via scanning directories -- we don't care.)
+        elif was_specified_by_user:
+            if kind is None:
+                raise ValueError(f"Unknown document type for {url}")
+            else:
+                raise NotImplementedError(
+                    f"Cannot currently validate {kind.name} files"
+                )
+        else:
+            # Not a doc type we recognise, and the user didn't specify it. Skip it.
+            continue
+
+        yield url, messages
 
 
-def _readable_doc_extension(path: Path):
+def _readable_doc_extension(path: str):
     """
-    >>> _readable_doc_extension(Path('something.json.gz'))
+    >>> _readable_doc_extension('something.json.gz')
     '.json.gz'
-    >>> _readable_doc_extension(Path('something.yaml'))
+    >>> _readable_doc_extension('something.yaml')
     '.yaml'
-    >>> _readable_doc_extension(Path('apple.odc-metadata.yaml.gz'))
+    >>> _readable_doc_extension('apple.odc-metadata.yaml.gz')
     '.yaml.gz'
-    >>> _readable_doc_extension(Path('/tmp/human.06.tall.yml'))
+    >>> _readable_doc_extension('/tmp/human.06.tall.yml')
     '.yml'
     >>> # Not a doc, even though it's compressed.
-    >>> _readable_doc_extension(Path('db_dump.gz'))
-    >>> _readable_doc_extension(Path('/tmp/nothing'))
+    >>> _readable_doc_extension('db_dump.gz')
+    >>> _readable_doc_extension('/tmp/nothing')
     """
     compression_formats = (".gz",)
     doc_formats = (
@@ -555,7 +564,7 @@ def _readable_doc_extension(path: Path):
     )
     suffix = "".join(
         s.lower()
-        for s in path.suffixes
+        for s in Path(path).suffixes
         if s.lower() in doc_formats + compression_formats
     )
     # If it's only compression, no doc format, it's not valid.
@@ -564,21 +573,47 @@ def _readable_doc_extension(path: Path):
     return suffix or None
 
 
-def expand_directories(
-    input_paths: Iterable[Path],
+def read_paths(
+    input_paths: Iterable[str],
+) -> Generator[Tuple[str, Union[Dict, str], bool], None, None]:
+    """
+    Read the given input paths, returning a URL, document, and whether
+    it was explicitly given by the user.
+
+    When a local directory is specified, inner readable docs are returned, but will
+    be marked as not explicitly specified.
+    """
+    for input_ in input_paths:
+        for uri, was_specified in expand_paths_as_uris([input_]):
+            try:
+                for full_uri, doc in read_documents(uri, uri=True):
+                    yield full_uri, doc, was_specified
+            except InvalidDocException as e:
+                if was_specified:
+                    raise
+                else:
+                    echo(e, err=True)
+
+
+def expand_paths_as_uris(
+    input_paths: Iterable[str],
 ) -> Generator[Tuple[Path, bool], None, None]:
     """
     For any paths that are directories, find inner documents that are known.
 
-    Returns Tuples: path, and whether it was specified explicitly by user.
+    Returns Tuples: path as a URL, and whether it was specified explicitly by user.
     """
-    for path in input_paths:
-        if path.is_dir():
-            for found_path in path.rglob("*"):
-                if _readable_doc_extension(found_path) is not None:
-                    yield found_path, False
+    for input_ in input_paths:
+        if is_url(input_):
+            yield input_, True
         else:
-            yield path, True
+            path = Path(input_).resolve()
+            if path.is_dir():
+                for found_path in path.rglob("*"):
+                    if _readable_doc_extension(found_path.as_uri()) is not None:
+                        yield found_path.as_uri(), False
+            else:
+                yield path.as_uri(), True
 
 
 def validate_eo3_doc(
@@ -591,25 +626,11 @@ def validate_eo3_doc(
     messages = []
 
     # TODO: follow ODC's match rules?
-    product = None
-    product_name = get_path(doc, ("product", "name"), default=None)
+
+    matched_product = None
 
     if products:
-        if len(products) == 1:
-            [product] = products.values()
-        elif product_name is not None:
-            product = products.get(product_name)
-
-        if product is None:
-            messages.append(
-                _warning(
-                    "unknown_product",
-                    "Cannot match dataset to product",
-                    hint=f"Nothing matches {product_name!r}"
-                    if product_name
-                    else "No product name in dataset (TODO: field matching)",
-                )
-            )
+        matched_product, messages = _match_product(doc, products)
     else:
         messages.append(
             ValidationMessage(
@@ -622,7 +643,7 @@ def validate_eo3_doc(
     messages.extend(
         validate_dataset(
             doc,
-            product_definition=product,
+            product_definition=matched_product,
             readable_location=location,
             thorough=thorough,
             expect_extra_measurements=expect_extra_measurements,
@@ -631,13 +652,127 @@ def validate_eo3_doc(
     return messages
 
 
+def _get_printable_differences(dict1: Dict, dict2: Dict):
+    """
+    Get a series of lines to print that show the reason that dict1 is not a superset of dict2
+    """
+    dict1 = dict(utils.flatten_dict(dict1))
+    dict2 = dict(utils.flatten_dict(dict2))
+
+    for path in dict2.keys():
+        v1, v2 = dict1.get(path), dict2.get(path)
+        if v1 != v2:
+            yield f"{path}: {v1!r} != {v2!r}"
+
+
+def _get_product_mismatch_reasons(dataset_doc: Dict, product_definition: Dict):
+    """
+    Which fields don't match the given dataset doc to a product definition?
+
+    Gives human-readable lines of text.
+    """
+    yield from _get_printable_differences(dataset_doc, product_definition["metadata"])
+
+
+def _match_product(
+    dataset_doc: Dict, product_definitions: Dict[str, Dict]
+) -> Tuple[Optional[Dict], List[ValidationMessage]]:
+    """Match the given dataset to a product definition"""
+
+    product = None
+
+    # EO3 datasets often put the product name directly inside.
+    specified_product_name = get_path(dataset_doc, ("product", "name"), default=None)
+    specified_product_name = specified_product_name or get_path(
+        dataset_doc, ("properties", "odc:product"), default=None
+    )
+
+    if specified_product_name and (specified_product_name in product_definitions):
+        product = product_definitions[specified_product_name]
+
+    matching_products = {
+        name: definition
+        for name, definition in product_definitions.items()
+        if changes.contains(dataset_doc, definition["metadata"])
+    }
+
+    # We we have nothing, give up!
+    if (not matching_products) and (not product):
+
+        # Find the product that most closely matches it, to helpfully show the differences!
+        closest_product_name = None
+        closest_differences = None
+        for name, definition in product_definitions.items():
+            diffs = tuple(_get_product_mismatch_reasons(dataset_doc, definition))
+            if (closest_differences is None) or len(diffs) < len(closest_differences):
+                closest_product_name = name
+                closest_differences = diffs
+
+        difference_hint = _differences_as_hint(closest_differences)
+        return None, [
+            _error(
+                "unknown_product",
+                "Dataset does not match the given products",
+                hint=f"Closest match is {closest_product_name}, with differences:"
+                f"\n{difference_hint}",
+            )
+        ]
+
+    messages = []
+
+    if specified_product_name not in matching_products:
+        if product:
+            difference_hint = _differences_as_hint(
+                _get_product_mismatch_reasons(dataset_doc, product)
+            )
+            messages.append(
+                _info(
+                    "strange_product_claim",
+                    f"Dataset claims to be product {specified_product_name!r}, but doesn't match its fields",
+                    hint=f"{difference_hint}",
+                )
+            )
+        else:
+            messages.append(
+                _info(
+                    "unknown_product_claim",
+                    f"Dataset claims to be product {specified_product_name!r}, but it wasn't supplied.",
+                )
+            )
+
+    if len(matching_products) > 1:
+        matching_names = ", ".join(matching_products.keys())
+        messages.append(
+            _error(
+                "product_match_clash",
+                "Multiple products match the given dataset",
+                hint=f"Maybe you need more fields in the 'metadata' section?\n"
+                f"Claims to be a {specified_product_name!r}, and matches {matching_names!r}"
+                if specified_product_name
+                else f"Maybe you need more fields in the 'metadata' section?\n"
+                f"Matches {matching_names!r}",
+            )
+        )
+        # (We wont pick one from the bunch here. Maybe they already matched one above to use in continuing validation.)
+
+    # Just like ODC, match rules will rule all. Even if their metadata has a "product_name" field.
+    if len(matching_products) == 1:
+        [product] = matching_products.values()
+
+    return product, messages
+
+
+def _differences_as_hint(product_diffs):
+    return indent("\n".join(product_diffs), prefix="\t")
+
+
 def _validate_stac_properties(dataset: DatasetDoc):
     for name, value in dataset.properties.items():
-        if name not in dataset.properties.KNOWN_STAC_PROPERTIES:
+        if name not in dataset.properties.KNOWN_PROPERTIES:
             yield _warning("unknown_property", f"Unknown stac property {name!r}")
 
         else:
-            normaliser = dataset.properties.KNOWN_STAC_PROPERTIES.get(name)
+            normaliser = dataset.properties.KNOWN_PROPERTIES.get(name)
             if normaliser and value is not None:
                 try:
                     normalised_value = normaliser(value)
@@ -645,6 +780,13 @@ def _validate_stac_properties(dataset: DatasetDoc):
                     if isinstance(normalised_value, tuple):
                         normalised_value = normalised_value[0]
 
+                    # It's okay for datetimes to be strings
+                    # .. since ODC's own loader does that.
+                    if isinstance(normalised_value, datetime) and isinstance(
+                        value, str
+                    ):
+                        value = ciso8601.parse_datetime(value)
+
                     # Special case for dates, as "no timezone" and "utc timezone" are treated identical.
                     if isinstance(value, datetime):
                         value = default_utc(value)
@@ -693,14 +835,15 @@ def _is_nan(v):
     return isinstance(v, float) and math.isnan(v)
 
 
-def _validate_geo(dataset: DatasetDoc):
+def _validate_geo(dataset: DatasetDoc, expect_geometry: bool = True):
     has_some_geo = _has_some_geo(dataset)
-    if not has_some_geo:
+    if not has_some_geo and expect_geometry:
         yield _info("non_geo", "No geo information in dataset")
         return
 
     if dataset.geometry is None:
-        yield _error("incomplete_geo", "Dataset has some geo fields but no geometry")
+        if expect_geometry:
+            yield _info("incomplete_geo", "Dataset has some geo fields but no geometry")
     elif not dataset.geometry.is_valid:
         yield _error(
             "invalid_geometry",
@@ -748,13 +891,14 @@ def _has_some_geo(dataset):
     help=__doc__
     + """
 Paths can be products, dataset documents, or directories to scan (for files matching
-names '*.odc-metadata.yaml' etc).
+names '*.odc-metadata.yaml' etc), either local or URLs.
 
-But each product must be specified before its datasets to be validated against them.
+Datasets are validated against matching products that have been scanned already, so specify
+products first, and datasets later, to ensure they can be matched.
 """
 )
 @click.version_option()
-@click.argument("paths", nargs=-1, type=PathPath(exists=True, readable=True))
+@click.argument("paths", nargs=-1)
 @click.option(
     "--warnings-as-errors",
     "-W",
@@ -774,6 +918,18 @@ But each product must be specified before its datasets to be validated against t
     help="Allow some dataset measurements to be missing from the product definition. "
     "This is (deliberately) allowed by ODC, but often a mistake. This flag disables the warning.",
 )
+@click.option(
+    "--explorer-url",
+    "explorer_url",
+    help="Use product definitions from the given Explorer URL to validate datasets. "
+    'Eg: "https://explorer.dea.ga.gov.au/"',
+)
+@click.option(
+    "--odc",
+    "use_datacube",
+    is_flag=True,
+    help="Use product definitions from datacube to validate datasets",
+)
 @click.option(
     "-q",
     "--quiet",
@@ -782,23 +938,35 @@ But each product must be specified before its datasets to be validated against t
     help="Only print problems, one per line",
 )
 def run(
-    paths: List[Path],
+    paths: List[str],
     strict_warnings,
     quiet,
     thorough: bool,
     expect_extra_measurements: bool,
+    explorer_url: str,
+    use_datacube: bool,
 ):
     validation_counts: Counter[Level] = collections.Counter()
     invalid_paths = 0
+    current_location = Path(".").resolve().as_uri() + "/"
+
+    product_definitions = _load_remote_product_definitions(use_datacube, explorer_url)
 
     s = {
         Level.info: dict(),
         Level.warning: dict(fg="yellow"),
         Level.error: dict(fg="red"),
     }
-    for path, path_index, messages in validate_paths(
-        paths, thorough=thorough, expect_extra_measurements=expect_extra_measurements
+
+    for url, messages in validate_paths(
+        paths,
+        thorough=thorough,
+        expect_extra_measurements=expect_extra_measurements,
+        product_definitions=product_definitions,
     ):
+        if url.startswith(current_location):
+            url = url[len(current_location) :]
+
         levels = collections.Counter(m.level for m in messages)
         is_invalid = levels[Level.error] > 0
         if strict_warnings:
@@ -809,8 +977,7 @@ def run(
             messages = [m for m in messages if m.level != Level.info]
 
         if messages or not quiet:
-            path_suffix = f" document {path_index+1}" if path_index else ""
-            secho(f"{bool_style(not is_invalid)} {path}{path_suffix}")
+            secho(f"{bool_style(not is_invalid)} {url}")
 
         if not messages:
             continue
@@ -826,7 +993,11 @@ def run(
                 f"\t{message.level.name[0].upper()} {displayable_code} {message.reason}"
             )
             if message.hint:
-                echo(f'\t\t({style("Hint")}: {message.hint})')
+                if "\n" in message.hint:
+                    echo("\t\tHint:")
+                    echo(indent(message.hint, "\t\t" + (" " * 5)))
+                else:
+                    echo(f'\t\t({style("Hint")}: {message.hint})')
 
     if not quiet:
         result = (
@@ -847,3 +1018,55 @@ def run(
             secho(f"{len(paths)} paths", err=True)
 
     sys.exit(invalid_paths)
+
+
+def _load_remote_product_definitions(
+    from_datacube: bool = False,
+    from_explorer_url: Optional[str] = None,
+) -> Dict[str, Dict]:
+
+    product_definitions = {}
+    # Load any remote products that were asked for.
+    if from_explorer_url:
+        for definition in _load_explorer_product_definitions(from_explorer_url):
+            product_definitions[definition["name"]] = definition
+
+    if from_datacube:
+        # The normal datacube environment variables can be used to choose alternative configs.
+        with Datacube(app="eo3-validate") as dc:
+            for product in dc.index.products.get_all():
+                product_definitions[product.name] = product.definition
+
+    secho(f"{len(product_definitions)} ODC products")
+    return product_definitions
+
+
+def _load_doc(url):
+    return list(load_documents(url))
+
+
+def _load_explorer_product_definitions(
+    explorer_url: str,
+    workers: int = 6,
+) -> Generator[Dict, None, None]:
+    """
+    Read all product yamls from the given Explorer instance,
+
+    eg: https://explorer.dea.ga.gov.au/products/ls5_fc_albers.odc-product.yaml
+    """
+    product_urls = [
+        urljoin(explorer_url, f"/products/{name.strip()}.odc-product.yaml")
+        for name in urlopen(urljoin(explorer_url, "products.txt"))
+        .read()
+        .decode("utf-8")
+        .split("\n")
+    ]
+    count = 0
+    with multiprocessing.Pool(workers) as pool:
+        for product_definitions in pool.imap_unordered(_load_doc, product_urls):
+            count += 1
+            echo(f"\r{count} Explorer products", nl=False)
+            yield from product_definitions
+        pool.close()
+        pool.join()
+    echo()
