@@ -10,8 +10,9 @@ import os
 import re
 import sys
 from datetime import timedelta, datetime
+from enum import Enum
 from pathlib import Path
-from typing import List, Sequence, Optional, Iterable, Any, Tuple, Dict, Mapping
+from typing import List, Sequence, Optional, Iterable, Any, Tuple, Dict
 from uuid import UUID
 
 import attr
@@ -27,9 +28,10 @@ from rasterio.enums import Resampling
 from eodatasets3 import serialise, utils, images, DatasetAssembler
 from eodatasets3.images import GridSpec
 from eodatasets3.model import DatasetDoc
+from eodatasets3.properties import Eo3Interface
 from eodatasets3.serialise import loads_yaml
 from eodatasets3.ui import bool_style
-from eodatasets3.utils import default_utc
+from eodatasets3.utils import default_utc, flatten_dict
 
 try:
     import h5py
@@ -57,6 +59,11 @@ FILENAME_TIF_BAND = re.compile(
 PRODUCT_SUITE_FROM_GRANULE = re.compile("(L1[GTPCS]{1,2})")
 
 
+class ProductMaturity(Enum):
+    provisional = "provisional"
+    stable = "stable"
+
+
 def _find_h5_paths(h5_obj: h5py.Group, dataset_class: str = "") -> List[str]:
     """
     Find all objects in a h5 of the given class, returning their path.
@@ -83,10 +90,11 @@ def _unpack_products(
     img_paths = _find_h5_paths(h5group, "IMAGE")
 
     for product in product_list:
-        with do(f"Starting {product}", heading=True):
+        with sub_product(product, p):
             for pathname in [
                 p for p in img_paths if "/{}/".format(product.upper()) in p
             ]:
+
                 with do(f"Path {pathname!r}"):
                     dataset = h5group[pathname]
                     band_name = utils.normalise_band_name(dataset.attrs["alias"])
@@ -102,13 +110,22 @@ def _unpack_products(
                 red, green, blue = _THUMBNAILS[product]
                 with do(f"Thumbnailing {product}"):
                     p.write_thumbnail(
-                        red, green, blue, kind=product, static_stretch=(1, 3000)
+                        red,
+                        green,
+                        blue,
+                        static_stretch=(1, 3000),
+                        # Because of our strange sub-products and filename standards, we want the
+                        # 'kind' to be included in the recorded thumbnail accessory metadata,
+                        # but not in the filename.
+                        # So we manually calculate a filename without the 'kind' field included.
+                        kind=product,
+                        path=p.names.thumbnail_filename(),
                     )
 
 
 def write_measurement_h5(
     p: DatasetAssembler,
-    name: str,
+    full_name: str,
     g: h5py.Dataset,
     overviews=images.DEFAULT_OVERVIEWS,
     overview_resampling=Resampling.nearest,
@@ -123,8 +140,8 @@ def write_measurement_h5(
     else:
         data = g
 
+    product_name, band_name = full_name.split(":")
     p.write_measurement_numpy(
-        name=name,
         array=data,
         grid_spec=images.GridSpec(
             shape=g.shape,
@@ -136,6 +153,12 @@ def write_measurement_h5(
         overview_resampling=overview_resampling,
         expand_valid_data=expand_valid_data,
         file_id=file_id,
+        # Because of our strange sub-products and filename standards, we want the
+        # product_name to be included in the recorded band metadata,
+        # but not in its filename.
+        # So we manually calculate a filename without the extra product name prefix.
+        name=full_name,
+        path=p.names.measurement_filename(band_name, "tif", file_id=file_id),
     )
 
 
@@ -301,6 +324,11 @@ def _create_contiguity(
             nodata=255,
             overviews=None,
             expand_valid_data=False,
+            # Because of our strange sub-products and filename standards, we want the
+            # 'oa_' prefix to be included in the recorded band metadata,
+            # but not in its filename.
+            # So we manually calculate a filename without the extra prefix.
+            path=p.names.measurement_filename(f"{product.lower()}_contiguity"),
         )
 
         # masking the timedelta_data with contiguity mask to get max and min timedelta within the NBAR product
@@ -346,6 +374,24 @@ def do(name: str, heading=False, **fields):
         secho("(done)")
 
 
+@contextlib.contextmanager
+def sub_product(name: str, p: Eo3Interface):
+    """
+    Set the product family temporarily within a block of code.
+
+    This is done for sub-products that WAGL contains, which have
+    a different 'family' in their filenames.
+    """
+    with do(f"Product {name}", heading=True):
+        original_family = p.product_family
+        # We delete first to show that we're deliberately changing the value (no 'overridding property" warning)
+        del p.product_family
+        p.product_family = name
+        yield
+        del p.product_family
+        p.product_family = original_family
+
+
 def _extract_reference_code(p: DatasetAssembler, granule: str) -> Optional[str]:
     matches = None
     if p.platform.startswith("landsat"):
@@ -534,6 +580,8 @@ def package_file(
 def package(
     out_directory: Path,
     granule: Granule,
+    *,
+    product_maturity: ProductMaturity = ProductMaturity.stable,
     included_products: Iterable[str] = DEFAULT_PRODUCTS,
     include_oa: bool = True,
     oa_resolution: Optional[Tuple[float, float]] = None,
@@ -582,6 +630,12 @@ def package(
                 processed=p.processed,
                 wagl_doc=wagl_doc,
             )
+
+            # We don't bother including product maturity if it's stable, for consistency with old datasets.
+            # Stable is the assumed default.
+            if product_maturity is not ProductMaturity.stable:
+                p.product_maturity = product_maturity
+
             if granule.source_level1_metadata is not None:
                 # For historical consistency: we want to use the instrument that the source L1 product
                 # came from, not the instruments reported from the WAGL doc.
@@ -617,48 +671,39 @@ def package(
             _unpack_products(p, included_products, granule_group)
 
             if include_oa:
-                with do("Starting OA", heading=True):
-                    _unpack_observation_attributes(
-                        p,
-                        included_products,
-                        granule_group,
-                        infer_datetime_range=p.platform.startswith("landsat"),
-                        oa_resolution=oa_resolution,
-                    )
-                if granule.fmask_image:
-                    with do(f"Writing fmask from {granule.fmask_image} "):
-                        p.write_measurement(
-                            "oa:fmask",
-                            granule.fmask_image,
-                            expand_valid_data=False,
-                            overview_resampling=Resampling.mode,
+                with sub_product("oa", p):
+                    with do("Starting OA", heading=True):
+                        _unpack_observation_attributes(
+                            p,
+                            included_products,
+                            granule_group,
+                            infer_datetime_range=p.platform.startswith("landsat"),
+                            oa_resolution=oa_resolution,
                         )
+                    if granule.fmask_image:
+                        with do(f"Writing fmask from {granule.fmask_image} "):
+                            p.write_measurement(
+                                "oa:fmask",
+                                granule.fmask_image,
+                                expand_valid_data=False,
+                                overview_resampling=Resampling.mode,
+                                # Because of our strange sub-products and filename standards, we want the
+                                # 'oa_' prefix to be included in the recorded band metadata,
+                                # but not in its filename.
+                                # So we manually calculate a filename without the extra prefix.
+                                path=p.names.measurement_filename("fmask"),
+                            )
 
             with do("Finishing package"):
                 return p.done()
 
 
-def _flatten_dict(d: Mapping, prefix=None, separator=".") -> Iterable[Tuple[str, Any]]:
-    """
-    >>> dict(_flatten_dict({'a' : 1, 'b' : {'inner' : 2},'c' : 3}))
-    {'a': 1, 'b.inner': 2, 'c': 3}
-    >>> dict(_flatten_dict({'a' : 1, 'b' : {'inner' : {'core' : 2}}}, prefix='outside', separator=':'))
-    {'outside:a': 1, 'outside:b:inner:core': 2}
-    """
-    for k, v in d.items():
-        name = f"{prefix}{separator}{k}" if prefix else k
-        if isinstance(v, Mapping):
-            yield from _flatten_dict(v, prefix=name, separator=separator)
-        else:
-            yield name, v
-
-
 def _read_gqa_doc(p: DatasetAssembler, doc: Dict):
     _take_software_versions(p, doc)
     p.extend_user_metadata("gqa", doc)
 
     # TODO: more of the GQA fields?
-    for k, v in _flatten_dict(doc["residual"], separator="_"):
+    for k, v in flatten_dict(doc["residual"], separator="_"):
         p.properties[f"gqa:{k}"] = v
 
 
